/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include "bresenham.h"

#ifndef BRESENHAM_BEAM_XY
#define BRESENHAM_BEAM_XY(X,Y)  bl->beam_xy (X,Y)
#endif // BRESENHAM_BEAM_XY

// http://de.wikipedia.org/wiki/Bresenham-Algorithmus

void bh_initline (bh_linedraw_t * bl,
                  void (*beam_xy) (vf_uint_t, vf_uint_t),
                  vf_upixel_t start, vf_upixel_t end,
                  uint8_t start_end)
{
    bl->pix = 0;
    
#ifndef BRESENHAM_BEAM_XY   
    bl->beam_xy = beam_xy;
#else
    (void) beam_xy;
#endif    

    bl->start = start;
    bl->end   = end;

    // Entfernung in beiden Dimensionen berechnen
    vf_spixel_t d =
    {
        .x = end.x - start.x,
        .y = end.y - start.y
    };

    // Vorzeichen des Inkrements bestimmen
    int8_t incx = 0;
    int8_t incy = 0;

    if (d.x)     incx = 1;
    if (d.y)     incy = 1;

    if (d.x < 0) d.x = -d.x, incx = -incx;
    if (d.y < 0) d.y = -d.y, incy = -incy;

    // dd. ist Diagonalschritt
    bl->ddx = incx;
    bl->ddy = incy;

    // Fehlerschritte schnell, langsam
    // el ist auch Anzahl zu zeichnender Pixel
    vf_int_t el;
    // pd. ist Parallelschritt

    // feststellen, welche Entfernung grer ist
    if (d.x > d.y)
        // x ist schnelle Richtung
        bl->pdx = incx, bl->pdy = 0,    bl->es = d.y, el = d.x;
    else
        // y ist schnelle Richtung
        bl->pdx = 0,    bl->pdy = incy, bl->es = d.x, el = d.y;

    vf_uint_t npix = el;

    if (! (start_end & BL_DRAW_END))
        npix--;

    bl->npix = npix;
    bl->el = el;
    bl->draw_start = start_end & BL_DRAW_START;
}

uint8_t bh_drawline (bh_linedraw_t * bl)
{
    // Momentane Koordinaten
    vf_upixel_t p;

    // Fehler
    vf_int_t err;

    // Fehlerschritte langsam
    vf_uint_t const el = bl->el;

    // Nummer des Pixels
    vf_uint_t pix = bl->pix;

    if (pix == 0)
    {
        // Start der Linie
        p   = bl->start;
        err = el / 2;

        if (! bl->draw_start)
        {
            if (++pix > bl->npix)
                goto end;
            else
                goto next_pixel;
        }
    }
    else
    {
        if (pix > bl->npix)
        {
        end:
            // Ende der Linie
            bl->pix = 0;
            return 1;
        }

        err = bl->err;
        p   = bl->p;
    next_pixel:

        // Aktualisierung Fehler
        err -= bl->es;
        if (err < 0)
        {
            // Fehler wieder nicht-negativ machen
            err += el;

            // Schritt in langsame Richtung, Diagonalschritt
            // dd. ist Diagonalschritt
            p.x += bl->ddx;
            p.y += bl->ddy;
        }
        else
        {
            // Schritt in schnelle Richtung, Parallelschritt
            // pd. ist Parallelschritt
            p.x += bl->pdx;
            p.y += bl->pdy;
        }
    }

    bl->pix = pix+1;
    bl->err = err;
    bl->p   = p;

    BRESENHAM_BEAM_XY (p.x, p.y);

    return 0;
}

void bh_initpoly (bh_polydraw_t * draw,
                  bh_sprite_t * sprite,
                  uint8_t num_sprites)
{
    draw->sprite          = sprite;
    draw->read_n_points = 2;
    draw->num_sprites     = num_sprites;
}

#include "compat/pgmspace.h"
//#include "reload.h"

uint8_t bh_drawpoly (bh_polydraw_t * draw)
{
    uint8_t read_n_points = draw->read_n_points;
    uint8_t flags = draw->flags;
    
    if (2 == read_n_points)
    {
        bh_sprite_t * line = draw->sprite;
        //RELOAD ("z", line);
        // 2 Punkte lesen: ein neuer Polygonzug fngt an
        // Besser von Hand kopieren als (implizites) memcpy
        // All diese Punkte werden mit gleichem Shift und Rotation
        // transformiert: P' = pos + rot*P
        uint8_t sprite_id = line->sprite_id;
        
        if (sprite_id == BH_SPRITE_NIL)
        {
            draw->sprite++;
            goto next_sprite;
        }
        uint8_t x = line->pos.x; draw->sprite_pos.x = x + 0x80;
        uint8_t y = line->pos.y; draw->sprite_pos.y = y + 0x80;
        draw->sprite_sincos   = _i_get_sincos8 (line->rot);
        draw->sprite_points = (const bh_polypoint_t*) pgm_read_addr (& draw->sprites[sprite_id]);
        draw->sprite++;
    }

    if (read_n_points)
    {
        // Nchsten Punkt des Polygonzuges 
        const bh_polypoint_t * pp = draw->sprite_points;
        
        vf_upixel_t start = draw->start;
        vf_upixel_t end;
        
        do
        {
            uint8_t px = pgm_read_byte_inc (pp);
            uint8_t py = pgm_read_byte_inc (pp);
            
            frotate8 (end.x, end.y, px, py, draw->sprite_sincos);
            
            end.x += draw->sprite_pos.x;
            end.y += draw->sprite_pos.y;
            
            flags = pgm_read_byte_inc (pp);
            
            if (flags & BL_SKIP)
            {
                BRESENHAM_BEAM_SKIP();
                read_n_points = 2;
            }
            
            if (2 == read_n_points)
                start = end;
                
        } while (--read_n_points);
    
        draw->sprite_points = pp;
        draw->start = end;
        draw->flags = flags;
        draw->read_n_points = 0;
        
        bh_initline (draw->linedraw, (void*) 0, start, end, flags);
    }
    
    if (bh_drawline (draw->linedraw))
    {
        uint8_t read_n_points = 1;
        
        if (draw->flags & BL_LIST_END)
        {
next_sprite:;
            uint8_t num_sprites = draw->num_sprites;
            
            if (--num_sprites == 0)
                return 1;
            
            draw->num_sprites = num_sprites;
            read_n_points = 2;
        }
            
        draw->read_n_points = read_n_points;
    }
    
    return 0;
}
