/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include "compat/interrupt.h"
#include "compat/pgmspace.h"

#include "attrib.h"
#include "main.h"

#include "rc5-codes.h"
#include "rc5-pd.h"
#include "ports.h"
#include "countdown.h"

#include "mem-check.h"
#include "panic.h"

#if defined DEBUG
#   include "mem-check.h"
#   include "soft-reset.h"
#   include "uart.h"
#   include "uput.h"
#else // !DEBUG
#   undef  PUT_P
#   define put_P(...)       (void) 0
#   define soft_reset(...)  (void) 0
#   define uart_getc_nowait(...) (-1)
#endif // DEBUG

#include "time.h"
#include "frame.h"
#include "sound.h"
#include "morse.h"

/*
   Folgende Module enthalten am Ende der C-Datei eine 
   Definition fr ein Objekt des Typs menuitem_t.
   In diesem Objekt wird das Verhalten fr den jeweiligen
   Men-Punkt festgelegt und Callbacks definiert:
   
   void * .data
        Zeiger auf Daten, die den folgenden Handlern beim Aufruf 
        bergeben wird.
        
   uint8_t bufsize
        Die Gre eines Anzeigepuffers. Es werden 3 dieser Puffer angelegt
        und Zeiger auf (in onRedraw) gefllte/freie Puffer an onRedraw 
        bzw. onPixel bergeben. Der Gesamtpuffer 'framebuf' in "frame.h" ist 
        eine Union, die fr das aktuelle Item ausgelegt und strukturiert ist.
        Weil immer nur ein Item aktiv ist, kann der Puffer eine Union sein;
        daher bentigt er nur so viel Platz wie das Dreifache des grten
        Einzelpuffers.
        
   uint8_t .redrawOn
        Eine Bitmaske, die festlegt wann onRedraw auszurufen 
        ist
        
   uint8_t .onKey (void * data, uint8_t code)
        Wird bei einem Tastendruck etc. mit dem entsprechenden Code
        aufgerufen. Falls der Handler TRUE zurckgibt und Bit REDRAW_KEY
        in .redrawOn gesetzt ist, wird nachfolgend onRedraw aufgerufen.
        
   void .onEnter (void * data)
        Wird nach der Aktivierung des Men-Items aufgerufen um 
        Initialisierungen zu ttigen.
        
   void .onPixel (void * data, void * buffer)
        Dieser Handler stellt die Pixel-Koordinaten fr die Anzeige
        bereit und kontrolliert die Strahl-Austastung.
        -- Dieser Handler luft innerhalb einer ISR.
        
        Koordinaten werden gesetzt durch Aufrufen von
           beam_xy (uint8_t x, uint8_t y) aus "frame.h"
        Strahlaustastung wird angefordert durch Aufruf von   
           beam_skip () aus "frame.h"
        Pro Aufruf mu genau ein Satz an Koordinaten erzeugt werden.
        Wird mehr als ein Satz pro Aufruf von onPixel erzeugt, dann 
        werden die vorherigen berschrieben. Wird kein Satz gesetzt, 
        dann wird der Handler erneut innerhalb der gleichen IRQ aufgerufen 
        (was aus Zeitgrnden vermieden werden sollte).
        Die Strahlaustastung erfolgt VOR der Ausgabe der Pixel-Koordinaten 
        auf die Rhre, auch wenn beam_skip() nach beam_xy() aufgerufen wird.
           
        Diese Routine wird im Takt von 48kHz aufgerufen, so da
        abzglich des ISR-Overheads nur einige hundert Ticks verbleiben
        um die Koordinaten zur Verfgung zu stellen. Bei komplizierten 
        Berechnungen knnen zur Beschleunigung im onRedraw-Handler 
        Vorberechnugen erfolgen.
        
        Der Aufruf von onPixel befindet sich in main.c:pixel_on_the_fly()
        innerhalb der OutputCompareA-ISR fr Timer #2.
        
        buffer ist ein Puffer, der nicht verndert wird whrend eines 
        Frames, d.h. solange Flag.FRAME_LOCKED gesetzt ist. Das ist notwendig,
        damit es nicht zu inkonsistenten Frames kommt, nachdem onRedraw einen
        neuen Frame fertiggestellt hat.
        Ist der Frame fertig gezeichnet, dann lscht onPixel Flag.FRAME_LOCKED 
        und es wird im nchsten Aufruf ein Puffer mit neuen Daten bergeben 
        (falls onRedraw einen fertig hat). 
        
        Falls keine Pufferung erforderlich ist, dann braucht Flag.FRAME_LOCKED
        nicht bedient zu werden und man setzt .onRedraw = 0.
        
   void .onRedraw (void * data, void * buffer)
        Vorberechnung/Aktualisierung der Grafikdaten fr die Anzeige. 
        Zweck dieses Handlers die Entlastung/Bescheunigung von onPixel.
        Der Handler schreibt seine Daten in den bereitgestellten Puffer.
        Es ist kein Problem, wenn onRedraw schneller arbeitet als ein
        Frame dauert, und es ist auch kein Problem, wenn die Berechnungen
        lnger dauern als ein Frame.
        
        Hat onRedraw keinen Frame vorberechnet, wenn onPixel mit einem neuen
        Frame beginnt, dann wird der alte nochmals angezeigt.
*/
#include "menu.h"
#include "text.h"
#include "uhr.h"
#include "testbild.h"
#include "snake.h"
#include "textmenu.h"
#include "schoner.h"
#include "dcf-oszi.h"

// #define put_P(...) (void) 0

vecfont_t vecfont;

menu_t menu;


/*
   Alle Text-Mens haben eine ID >= ITEMIDS = ITEMID_TEXTMENU.
   menu_activate_item bildet all diese IDs auf ITEMID_TEXTMENU ab,
   und die darzustellenden Texte sind in textmenu.c:textmenus[]
   abgelegt: Fr jede ID >= ITEMIDS gibt es dort einen Eintrag.
*/
static const menuitem_t * menuinit[] PROGMEM = 
{
    [ITEMID_UHR]       = & menuitem_uhr,
    [ITEMID_TEXT]      = & menuitem_text,
    [ITEMID_TEST]      = & menuitem_testbild,
    [ITEMID_SNAKE]     = & menuitem_snake,
    [ITEMID_SCHONER]   = & menuitem_schoner,
    [ITEMID_DCFOSZI]   = & menuitem_dcfoszi,
    [ITEMID_ASTEROIDS] = & menuitem_asteroids,
    
    [ITEMID_TEXTMENU]  = & menuitem_textmenu
};

void 
menu_activate_item (menu_t * m, uint8_t id)
{
    if (id >= ITEMIDS)
    {
        // Alle Items >= ITEMIDS sind Text-Mens und tun alle das
        // gleiche: Sie stellen Alternativen zur Auswahl und leiten
        // beim Anclicken weiter. (textmenu.c)
        uint8_t menu_id = id - ITEMIDS;
        m->menu_id = menu_id;
        m->item_count = pgm_read_byte (& textmenus[menu_id].item_count);
        id = ITEMIDS;
    }
    m->item_id = id;
    
    uint8_t sreg = SREG;
    cli();

    // Die Infos fr das neue MenuItem stehen im Flash...
    menuitem_t * pitem = (menuitem_t*) pgm_read_word (& menuinit[id]);
    
    // ... und werden nach menu kopiert, um besser zugreifen zu knnen
    memcpy_P (& m->item, pitem, sizeof (menuitem_t));
    
    // Alles initialisieren fr ein neues Item:
    // ...evtl. Sperrung des Strahls aufheben
    FLAGS &= ~FRAME_LOCKED;
    // ... vekfont zurcksetzen (vielleicht war er nicht fertig)
    vecfont_reset (&vecfont);
    // ... Frame-Puffer initialisieren
    frame_bufinit (&frame, menu.item.bufsize);
    
    SREG = sreg;

    // Menu-Item initialisieren (bzw. Merker setzen, da onRedraw
    // eine Erstinitialisierung vornimmt. Das ist zB in snake erforderlich,
    // weil es Daten zwischen einzelnen Frame-Puffern kopiert.)
    ((onEnter_t) m->item.onEnter) (m->item.data);
}

static inline void 
reset (void)
{
    sys.bootlevel = 0;//BOOT_HEATING;
    SET (PORT_Z);
    CLR (PORT_POWER);
    wait_10ms (50);
    soft_reset();
}

static void
handle_bootlevel (void)
{
    uint8_t blevel = sys.bootlevel;
    
    if (blevel >= BOOT_DONE  ||  count.sec.boot)
        return;

    // Dieser Boot-Level ist abgeschlossen, 
    // nchster fngt an
    uint8_t dauer = 0;
    uint8_t buch = 0;
    
    sys.bootlevel = ++blevel;
    
    if (BOOT_HEATING == blevel)
    {
        dauer = (sys.mcusr & (1 << PORF)) ? 10 : 7;
        buch = _I_;
    }
        
    if (BOOT_POWERING == blevel)
    {
        dauer = 3;
        
        SET (PORT_POWER);

        put_P ("Power ON!\n");
        buch = _S_;
    }
        
    if (BOOT_DONE != blevel)
        count.sec.boot = dauer;
    else
        buch = _H_;
        
    if (buch)
        morse (buch);
}

/*
   Die Hauptschleife:
   1) Boot-Level anpassen (handle_bootlevel)
  
   2) Anzeige-Puffer neu fllen falls erforderlich
      --> redrawOn legt fest, wann ein Neufllen notwendig ist, zB:
          -- alle 10ms
          -- bei Taste/Aktion
          -- nach Men-Umschaltung
      --> onRedraw-Handler
      
   3) Eingaben einlesen (UART, IR-Fernbedienung, ...
   
   4) Eingabe auswerten
   4a) Men-Umschaltung? --> menu_activate_item
   4b) Taste/Aktion?     --> onKey-Handler
  
*/
void 
menu_main (menu_t * m)
{
	rc5_init (0xff);
    uint8_t rc5_code = 0xff;
    
    menu_activate_item (m, ITEMID_TEST);
    
    frame.schoner_timeout = 255;
    count.sec.schoner = frame.schoner_timeout;
    
    uint8_t redraw = REDRAW_ENTER;
    
    while (1)
    {
        handle_bootlevel ();
        
        if (redraw && m->item.onRedraw)
        {
            m->redraw = redraw;
            ((onRedraw_t) m->item.onRedraw) (m->item.data, frame.buffers.tmp);
            
            frame_activate (&frame);
        }
        ///////////////////////////////////
        int ic = uart_getc_nowait ();
        
        redraw = 0;
        
        if (ic >= 0)
        {
            uint8_t c = ic;
            if (c == '\e')          reset();
/*            
            else if (c == 'P')      
            {
                count.sec.power_startup = 3;
                SET (PORT_POWER);
                put_P ("Set Power on\n");
            }
            else if (c == 'p')
            {
                CLR (PORT_POWER);
                put_P ("Set Power off\n"); 
            }
*/
#if 0
            else if (c == '?')
            {
                put_P ("\nframe.valid         = %d\n", IS_FLAG (BEAM_VALID)); 
                put_P ("bootlevel            = %d\n", sys.bootlevel); 
                put_P ("sec.schoner          = %d\n", count.sec.schoner); 
                put_P ("stat.loop            = %d\n", sys.stat.loop); 
            }
#endif            
            else if (c == 'U')      
                put_P ("Mem free = %d\n", get_mem_unused());
            else
            {
                if (m->item.onKey)
                    redraw |= ((onKey_t) m->item.onKey) (m->item.data, 0x80 | c);
            }
            
            //count.sec.schoner = frame.schoner_timeout;
        } // uart
        ///////////////////////////////////
        if (rc5_decode())
        {
            count.sec.schoner = frame.schoner_timeout;
            rc5_code = rc5.code;
        }
        
//        uint8_t code = rc5_code;
        
        if (rc5_code != 0xff)
        {
            if (rc5_code <= ITEMIDS)
            {
                put_P ("\nITEM -- %d --\n", rc5_code);
                {
                    uint8_t id = rc5_code;
                    if (id == 0) id = ITEMID_TEXTMENU;
                    else id--;
                    menu_activate_item (m, id);
                }
                redraw |= REDRAW_ENTER;
            }
            else if (rc5_code == RC5_OFF) 
                reset();
            else if (rc5_code == RC5_MUTE)      
            {
#warning FIXME RC5_MUTE
/*
                TOGGLE (PORT_POWER);
                
                if (N_TO_PORT (PORT_POWER) & ID_TO_BIT(PORT_POWER))
                {
                    count.sec.power_startup = 2;
                
                    do
                    {
                    } while (count.sec.power_startup);
                }
                */
            }
            else if (m->item.onKey)
            {
                //if (5 == code)  sys.debug = 0;
                //if (6 == code)  sys.debug = 1;

                if (((onKey_t) m->item.onKey) (m->item.data, rc5_code))
                    redraw |= REDRAW_KEY;
                    
  //              put_P ("RC5 = %d\n", rc5_code);
            }
            rc5_code = 0xff;
        } // rc5_code
        
        if (0) {
            if (get_mem_unused () < 50)
                panic (_X_);
        }

        ///////////////////////////////////
        {/*
            if (count.ms10.blink >= 90)
                SET (PORT_LED);
            else
                CLR (PORT_LED);
                */
            if (!count.ms10.blink)
            {
                timeid_add (&time, 1);
                count.ms10.blink = 100;
            }    
        }    
       
        ///////////////////////////////////
        {
            uint8_t redraw_on = m->item.redrawOn;
            
            if ((redraw_on & REDRAW_1_SECOND)
                && time.changed.second)
            {
                redraw |= REDRAW_1_SECOND;
                time.changed.second = 0;
            }
            
            if ((redraw_on & REDRAW_10_MILLI_SECONDS)
                && m->tick_10ms)
            {
                cli();
                m->tick_10ms = 0;
                redraw |= REDRAW_10_MILLI_SECONDS;
                sei();
            }
            
            if ((redraw_on & REDRAW_1_MILLI_SECOND)
                && m->tick_1ms)
            {
                cli();
                m->tick_1ms = 0;
                redraw |= REDRAW_1_MILLI_SECOND;
                sei();
            }
            
            if ((redraw_on & REDRAW_DCF)
                && m->dcf_bit >= 0)
            {
                cli();
                uint8_t dcf_code = 0x80 | m->dcf_bit;
                m->dcf_bit = -1;
                sei();
                if (((onKey_t) m->item.onKey) (m->item.data, dcf_code))
                    redraw |= REDRAW_DCF;
            }
        }
    } // while (1)
}
