/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include <string.h>

#include "snake.h"

#define SNAKE_PIXEL_PER_FRAME 1200

#include "compat/interrupt.h"
#include "compat/pgmspace.h"
#include "reload.h"
#include "frame.h"

#if defined  __AVR__
    #include "morse.h"
    #include "sound.h"
/*    
    #include "uput.h"
    // <stdio.h>
    #define printf(...) (void) 0
    #define sprintf(STR,FMT,X...) put_P ("%@" FMT, STR , ##X)
*/    
#else 
    #include "snake-pc-compat.h"
#endif // AVR, 386

#include "rc5-codes.h"
#include "parith-15.h"
#include "main.h"

typedef struct
{
    upixel8_t p;
    uint8_t dir;
    uint8_t len;
} snake_vector_t;

snake_t snake;

static int8_t snake_umlauf (snake_frame_t*, upixel8_t);
static void snake_move (snake_t*, snake_frame_t*);

static inline uint8_t snake_check_line (snake_vector_t, upixel8_t);
static uint8_t snake_point_on_snake (snake_frame_t * sb, uint8_t i0, upixel8_t q);

static char* 
ushort_to_string (char *str, uint16_t n)
{
    if (!n)
    {
        *str++ = '0';
        *str = '\0';
        return str;
    }
    
    uint16_t p = 10000;
    uint8_t not0 = 0;

    do 
    {
        char c = '0';
        
        while (n >= p)
            not0 = 1, c++, n -= p;

        if (not0)
            *str++ = c;
            
    } while (p /= 10);
    
    *str = '\0';
    return str;
}

// Lnge der Schlang beim Start
#define START_LENGTH 20

// Tastendruck/Kommando
static uint8_t 
onKey_snake (snake_t * s, uint8_t code)
{
    uint8_t dir = s->dir;

    if (RC5_LEFT  == code)   dir = SN_W;
    if (RC5_RIGHT == code)   dir = SN_E;
    if (RC5_UP2   == code)   dir = SN_N;
    if (RC5_DOWN2 == code)   dir = SN_S;
    if (RC5_UP    == code)   dir = SN_N;
    if (RC5_DOWN  == code)   dir = SN_S;
    
    uint8_t level = s->game.level;

    if (RC5_ROT  == code)    level = -1;
    if (RC5_GRUN == code)    level = -2;
    if (RC5_GELB == code)    level = -3;
    if (RC5_BLAU == code)    level = -4;
    
    s->game.level = level;
    
//    if (RC5_11 == code)      sys.debug = 10;
//    if (RC5_12 == code)      s->game.tick = 1;

    // Keine 180-Wende: Neue Richtung erst spter verarbeiten
    snake_frame_t *sb = (snake_frame_t*) frame.buffers.isr_new;

    if (dir != sb->body.seg[sb->body.start].dir)
        s->new_dir = dir & 3;
    
    return 1;
}

// Menu-Punkt "Snake" aktivieren
static void 
onEnter_snake (snake_t * s)
{
    s->game.over = 0;
    s->game.score = 0;
    
    s->game.level = -1;
}

static upixel8_t 
snake_random_pixel (void)
{
    uint8_t x = prandom();
    uint8_t y = prandom();

    return (upixel8_t) { x &= ~(SNAKE_RASTER-1), 
                         y &= ~(SNAKE_RASTER-1)};
}

static void 
snake_game_action (snake_t * s, snake_frame_t * sb)
{
    snake_move (s, sb);
    
    if (s->game.over)
    {
        s->scroll_y++;
        return;
    }
    
    if (1 != snake_umlauf (sb, sb->head)
        || snake_point_on_snake (sb, 1, sb->head))
    {
        s->level.delay  = 15;
        s->scroll_y     = 0;
        s->game.over    = 1;
        s->changed.text = 1;
        sound (1);
    }
}

/* Das komplette Spielgeschehen wird immer vom gerade angezeigten
   Teilpuffer in den aktuellen, zu berechnenden Puffer kopiert.
   Auf diese Weise spart man das Anlegen eines vierten Puffers,
   der das Spielgeschehen enthlt.
   Allerdings ist beim Kopieren zu beachten, da onPixel einige
   Daten verndert hat.
*/   
static void 
onRedraw_snake (snake_t * s, snake_frame_t * sb)
{
    int8_t level = s->game.level;
    
    if (level > 0)
        // Hier besser als memcpy
        *sb = *((snake_frame_t*) frame.buffers.isr_new);
    else
    {
        s->game.over = 0;
        snake_set_level (s, -level, sb);
    }

    // Wird gebraucht zur Umlauf-Berechnung in snake_read_wall_seg
    sb->level = s->game.level;
    
    if (menu.redraw & REDRAW_1_MILLI_SECOND)
    {
        // tempus fugit
        uint8_t ticks = s->game.ticks;
        
        if (++ticks >= s->level.delay)
        {
            snake_game_action (s, sb);
            ticks = 0;
        }
        s->game.ticks = ticks;
    
        if (s->seek_food)
        {
            // Neues Futter platzieren
            sb->food = snake_random_pixel(); 
            // Futter muss innerhalb der Umrandung sein und
            // soll nicht auf der Schlange liegen
            if (1 != snake_umlauf (sb, sb->food)
                || snake_point_on_snake (sb, 0, sb->food))
                // Bedingungen nicht erfllt: Futter ungltig. 
                // Nchster Versuch 1ms spter...
                sb->food.x = 0;
            else
                // Bedingungen erfllt: Suche beendet
                s->seek_food = 0;
        }
    }
    
    // Werte, die onPixel (und Umlauf-Berechnung) verndert(e) ...
    sb->draw.elem  = 0xf9;
    sb->draw.len   = 0;
    sb->draw.ntext = 0;
    
    if (s->game.over)
    {
        sb->food.x = 0;
        sb->ntexte = 3;
    }
    else
    {
        sb->ntexte = 2;
    }
    
    if (s->changed.text)
    {
        s->changed.text = 0;
        sb->str[0] = s->game.level+'0';
        sb->str[1] = '\n';
        //sprintf (sb->str+2, "%d", s->game.score);
        ushort_to_string (sb->str+2, s->game.score);
    }
}

// Schwierigkeitsgrad einstellen
void 
snake_set_level (snake_t * s, uint8_t level, snake_frame_t * sb)
{
    //FLAGS &= ~BEAM_ISR_NEW;
    memcpy_P (& s->level, & snake_levels[level-1] , sizeof (snake_level_t));
    s->game.level = level;
    s->game.ticks = 0;
    s->dir = prandom() & 3;
    s->new_dir = s->dir;
    s->grow = 0;
    // Futter ist noch nicht gefunden
    s->seek_food = 1;
    s->changed.text  = 1;
    
    // Schlong hat 1 Segment
    sb->dir = s->dir;
    sb->body.seg[0].dir = 2 ^ s->dir;
    sb->body.seg[0].len = START_LENGTH;
    sb->head = (upixel8_t) {160, 160};
    sb->body.start = 0;
    sb->body.nseg  = 1;
    // Futter ungltig
    sb->food.x = 0;
}

// Schlange einen Pixel weiter
void 
snake_move (snake_t * s, snake_frame_t * sb)
{
    uint8_t start = sb->body.start;
    uint8_t nseg  = sb->body.nseg;

    if (s->game.over)
    {
        if (!nseg)
            return;
        goto _schrumpf;
    }

    // Kopf 1 weiter
    uint8_t x = sb->head.x;
    uint8_t y = sb->head.y;

    uint8_t dir  = s->dir;
    uint8_t mask = SNAKE_RASTER-1;

    // Falls Kopf auf Rasterpunkt: ggf. neue Richtung bernehmen
    if (s->new_dir != 0xff
        && (x & mask) == 0 
        && (y & mask) == 0)
    {
        s->dir = sb->dir = dir = s->new_dir;
        s->new_dir = 0xff;
    }

    // Schritt in die Richtung
    if (dir & 1)    y += dir - 2;
    else            x += dir - 1;

    sb->head.x = x;
    sb->head.y = y;

    // Mampf?
    if (x == sb->food.x  &&  y == sb->food.y)
    {
        // Mampf!
        s->grow = s->level.grow;
        s->game.score += s->level.grow;
        s->changed.text = 1;

        // neues Futter suchen gehen
        s->seek_food = 1;
        sound (0);
    }

    // Erstes und letztes Segment anpassen
    uint8_t dirs = 2 ^ dir;

    if (dirs != sb->body.seg[start].dir)
    {
        // Richtung wurde gendert -> ein neues Segment vorhngen
        nseg++;

        // start-- (mod SEG_MAX)

        if (--start == 0xff)
            start = SEG_MAX-1;

        // Start-Index hat sich gendert
        sb->body.start = start;
        // Neues Segment initialisieren
        sb->body.seg[start].len = 0;
        sb->body.seg[start].dir = dirs;
    }

    // Erstes Segment 1 lnger
    sb->body.seg[start].len += 1;

    if (nseg > SEG_MAX)
        // zu viele Segmente: abschneiden
        nseg = SEG_MAX;
    else
    {
        // Noch wachsen?
        if (s->grow)
            s->grow--;
        else
        {
_schrumpf:;
            // Nicht mehr wachsen: letztes Segment 1 krzer
            // Ende 1 krzen
            // Index des letzten Segments (mod SEG_MAX)
            uint8_t idx  = start + nseg-1;
            if (idx >= SEG_MAX)
                idx -= SEG_MAX;

            uint8_t len = sb->body.seg[idx].len - 1;

            if (len)
                // Segment 1 krzer
                sb->body.seg[idx].len = len;
            else
                // Segment ist leer --> letztes Segment entfernen
                nseg--;
        }
    }

    if (sb->body.nseg != nseg)
    {
        sb->body.nseg = nseg;
        s->changed.text = 1;
    }
}

int8_t 
snake_umlauf (snake_frame_t * sb, upixel8_t q)
{
    int8_t umlauf = 0;
    
    sb->draw.wall = snake_wall;
    
    while (1)
    {
        snake_read_wall_seg (sb);
        if (sb->draw.elem == 0xfd)
            break;
            
        // Q: Testpunkt
        // P0/P1: Beginn/Ende des Mauerelements
        uint8_t x0  = sb->draw.x;
        uint8_t y0  = sb->draw.y;
        
        snake_vector_t vec = { {x0, y0}, .dir=sb->draw.dir, .len=sb->draw.len};
        
        int8_t check = (int8_t) snake_check_line (vec, q);
        
        if (2 == check)
            return 2;
        
        umlauf -= check;
        
        uint8_t dir = sb->draw.dir;    
        uint8_t len = sb->draw.len;
        
        // Schritt in die Richtung
        if (dir < 2)        len = -len;
        if (dir & 1)        y0 += len;
        else                x0 += len;
        
        sb->draw.x = x0;
        sb->draw.y = y0;
    }
    
    return umlauf;    
}

uint8_t 
snake_point_on_snake (snake_frame_t * sb, uint8_t i0, upixel8_t q)
{
    uint8_t nseg = sb->body.nseg;

    snake_vector_t line;
    line.p = sb->head;
    
    for (uint8_t i=0; i < nseg; i++)
    {
        uint8_t idx    = i + sb->body.start;
        
        if (idx >= SEG_MAX)
            idx -= SEG_MAX;
            
        line.len = sb->body.seg[idx].len;
        line.dir = sb->body.seg[idx].dir;
       
        if (i >= i0 
            && 2 == snake_check_line (line, q))
            return 1;
            
        // Schritt in die Richtung
        if (line.dir < 2)        line.len = -line.len;
        if (line.dir & 1)        line.p.y += line.len;
        else                     line.p.x += line.len;
    }
    
    return 0;
}

// 2: Punkt liegt auf dem Segment (inclusive Endpunkte)
// 0, 1, -1 (0xff): Beitrag dieser Linie zum Umlauf um q 
uint8_t 
snake_check_line (snake_vector_t vec, upixel8_t q)
{
    // ka darf nicht 0 sein falls kb im Intervall [k0,k1]
    uint8_t ka, kb;
    uint8_t k0, k1;

    if (vec.dir & 1)
    {
        // Linie in y-Richtung (NORD || SD)
        ka = q.x - vec.p.x;
        kb = q.y;
        k0 = vec.p.y;
    }
    else
    {
        // Linie in x-Richtung (OST || WEST)
        ka = q.y - vec.p.y;
        kb = q.x;
        k0 = vec.p.x;
    }

    k1 = k0;

    // Intervallgrenzen
    if (vec.dir >= 2)
        // NORD || OST
        k1 += vec.len;
    else
        // WEST || SD
        k0 -= vec.len;
    
    // Punkt auf dem Linienstck?
    if (ka == 0
        && kb >= k0 && kb <= k1)
        // Ja
        return 2;

    // Nein
    // Beitrag zum Umlauf berechnen
    if (vec.dir & 1)
        // NORD || SD
        return 0;
        
    // Teststrahl geht nach oben, d.h. Linie muss 
    // oberhalb des Testpunktes sein
    if (vec.p.y > q.y
        // Teststrahl trifft Mauerstck (x0, x1]
        && kb > k0 && kb <= k1)
        // WEST: -1
        // Ost:  +1
        return vec.dir -1;
        
    return 0;
}


// Decodiert die Spielfeld-Grenze gem snake_wall[]
// REENTRANT
// return skip
// Liest inkrementell ein Mauerelement.
//      .draw.wall:     Zeiger auf zu lesendes Byte von snake_wall[]
//      .draw.level:    Spiel-Level
// Setzt
//      .draw.dir:      die Richtung
//      .draw.x:        Koodinaten
//      .draw.y:
//      .draw.len:      die Lnge in Einheiten von 2 ** SNAKE_WALL_SKIP
//                      .draw.len = 0 beendet die aktuelle Teilmauer
//      .draw.elem:     setzt Element-No (fr Ausgabe). Alle Mauern sind fertig
//                      wenn .elem = 0xfd
// return:              .0: Skip: Springt der Strahl? Nur fr Ausgabe

uint8_t 
snake_read_wall_seg (snake_frame_t *sb)
{
    uint8_t dir_len, dir, len, dir_0;
    uint8_t sb_level, sb_level_40;
    const uint8_t * wall = sb->draw.wall;

    RELOAD ("y", sb);
    
    sb->draw.elem = 0xfd-1;
    sb_level    = sb->level;
    sb_level_40 = sb_level + 0x40;
    dir_0 = ((unsigned int) wall) ^ ((unsigned int) snake_wall);

 _next_seg:;

    uint8_t first = 0;
    dir_len  = pgm_read_byte_inc (wall);

    if (dir_len == SNAKE_WALL_GOTO)
    {
        first = 1;
        sb->draw.x = pgm_read_byte_inc (wall);
        sb->draw.y = pgm_read_byte_inc (wall);
        dir_len   = pgm_read_byte_inc (wall);
        dir_0++;
    }
    dir_0++;
    
    // dir.0: Weil das erste Element nach Osten geht, ergibt sich
    //        dir.0 aus Offset.0 in snake_wall[].
    dir = dir_0 & 1;
    // dir.1: Steht in Bit 0
    if (dir_len & 1)
        dir |= 2;

    uint8_t level = dir_len >> 1;
    level &= 0x43; // 0b01000011

    uint8_t level_ok = sb_level_40 <= level
        || sb_level > level;

    // level wegmaskieren
    dir_len &= ~0x86; // 0b01111001

    if (dir_len == 1)
    {
        // Noch eine Mauer.
        // Mauer in diesem Level aktiv?
        if (level_ok)
            // Ja.
            goto _next_seg;
            
        // Nein --> len = 0 beendet Mauer-Ausgabe
        dir_len = 0;
    }

    // Lnge extrahieren
    len = dir_len & 0x78; // b01111000

    // Segment in diesem Level aktiv?
    if (len && !level_ok)
        // Nein: nchstes Segment einlesen
        goto _next_seg;

    sb->draw.wall = wall;
    sb->draw.dir = dir;
    sb->draw.len = len;
    
    if (!len)
    {
        // Alle Mauern sind fertig
        sb->draw.elem = 0xfd;
        return 1;
    }
    
    return first;
}

// Menu-Punkt "Snake"
const menuitem_t menuitem_snake PROGMEM =
{
    .onPixel       = onPixel_snake,
    .onEnter       = onEnter_snake,
    .onRedraw      = onRedraw_snake,
    .onKey         = onKey_snake,
    .bufsize       = sizeof (snake_frame_t),
    .redrawOn      = REDRAW_1_MILLI_SECOND,
    .data          = & snake
};
