/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#ifndef _SNAKE_H_
#define _SNAKE_H_

#include <inttypes.h>

enum
{
    SN_HEAD, // Kopf
    SN_FOOD  // Futter
};

#define SEG_MAX 25
#define SNAKE_RASTER 8
    
typedef struct
{
    // Verzoegerung zwischen Aufrufen von onTimeChanged
    uint8_t delay;
    // Soviel waechst das Vieh wenns was frisst
    uint8_t grow;
} snake_level_t;

#include "pixel.h"

typedef struct 
{
    uint8_t dir;
    
    // einzelne Punkte: Kopf, Futter, ...
    union
    {
        struct
        {
            upixel8_t head;
            upixel8_t food;
        };
        upixel8_t p[2];
    };
    
    // 0..3 (snake.game.level -1) zum Zeichnen der Mauer, etc.
    uint8_t level;
    
    // Anzahl zu zeichender Texte
    uint8_t ntexte;

    // Die Daten in dieser Struktur sind "lokale" Variablen in onPixel_snake
    // und werden dirt verndert. Beim Erzeugen eines neuen Frames aus .isr_new
    // in onRedraw_snake sind sie daher ungltig, weil nicht bekannt ist, zu welchem
    // Zeitpunkt die Kopie erstellt wird. Daher mssen Variablen wie .elem dort
    // wieder initialisiert werden.
    struct
    {
        // momentan gezeichnetes Element
        // -2, -1: head
        // 0..nseg-1: body
        uint8_t elem;
        // Anzahl noch zu zeichnende Pixel des aktuellen Elements
        uint8_t len;
        // momentane Zeichenrichtung
        uint8_t dir;
        // Momentane Strahlposition
        uint8_t x, y;
        // Zeiger auf Mauer-Segment in snake_read_wall_seg()
        const uint8_t * wall;
        // Nummer des gerade gezeichneten Textes
        uint8_t ntext;
    } draw;
    
    // Krper
    struct
    {
        // Index des ersten Segments
        uint8_t start;
        // Anzahl Segmente
        uint8_t nseg;
        // Segmente
        // Analog wie in der Mauer snake_wall[] wird die Richtung in 
        // nur in Bit.0 codiert.
        struct
        {
            // Richtung + Lnge
            uint8_t dir, len;
        } seg[SEG_MAX];
    } body;
    
    // String fr Level+Score
    char str[15];
} snake_frame_t;

typedef struct snake_t
{
    // Kriech-Richtung
    uint8_t dir;
    // Neue Richtung angefordert
    uint8_t new_dir;
    
    // Futter lsst die Schlong wachsen
    uint8_t grow;

    // Wird gerade ein neuer Platz fuers Futter gesucht??
    uint8_t seek_food;
    
    // # zu zeichnender Texte
    uint8_t ntexte;
    
    // # y-Pos fuer "Game Over"
    uint8_t scroll_y;
    
    struct
    {
        uint8_t text;
    } changed;
    
    struct
    {
        // Eingestellter Level: 1..4
        uint8_t level;
        // Erreichte Punktzahl
        uint16_t score;
        // Spiel ist aus :-(
        uint8_t over;
        // Zeitraster
        uint8_t ticks;
    } game;
    
    snake_level_t level;
   
} snake_t;

extern snake_t snake;
extern void snake_set_level (snake_t*, uint8_t, snake_frame_t*);
extern void onPixel_snake (snake_t*, snake_frame_t*);

#define SNAKE_PIXEL_PER_FRAME 1200
#define SNAKE_WALL_SKIP 2

// Richtungen
// NORD ^ SD  = 2
// WEST ^ OST  = 2
// ++ dreht nach links  (math. pos.)
// -- dreht nach rechts (math. neg.)
#define SN_N 3 // Nord
#define SN_S 1 // Sd
#define SN_W 0 // West
#define SN_E 2 // Ost

typedef struct 
{
    upixel8_t start;
    uint8_t zoom;
    uint8_t text_in_flash;
    const char* pstr;
} snake_text_init_t;

extern const uint8_t snake_wall[];
extern const snake_level_t snake_levels[];
extern const snake_text_init_t snake_text_init[];
extern const pixel8_t snake_head_dxy[];

#define SNAKE_RASTER_LOG 3

#define SNAKE_VAL(DIR,LEN,LEVEL) (0                           \
        | (/*0b00000001*/ 0x01 & (DIR >> 1))            \
        | (/*0b01111000*/ 0x78 & (LEN << SNAKE_RASTER_LOG))   \
        | (/*0b10000110*/ 0x86 & ((LEVEL) << 1))  )

#define SNAKE_WALL_GOTO         SNAKE_VAL(2,0,0x40)
#define SNAKE_WALL_NEXT(LEVEL)  SNAKE_VAL(2,0,LEVEL-1)
#define SNAKE_WALL_END  0

// return skip
extern uint8_t snake_read_wall_seg (snake_frame_t *s);

#include "menu.h"
extern const menuitem_t menuitem_snake;

#endif /* _SNAKE_H_ */
