/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include "compat/interrupt.h"
#include "compat/pgmspace.h"
#include "attrib.h"

#include "main.h"
#include "wire.h"
#include "sintab.h"

#include "frame.h"
#include "rc5-codes.h"

#include "panic.h"

#include "testbild.h"

typedef struct test_t
{
    // Welches Testbild?: 0..3
    uint8_t id;
    
    // Laufvariablen
    uint16_t i, i2, ity;
    uint8_t k, j;
    uint8_t x0, y0;
    uint8_t x, y;
    uint8_t size, q,a;
    uint8_t mode;
    uint16_t inc[4];
    uint8_t cur;
    uint8_t radius[4];
    int16_t shift[4];
} test_t;

test_t test = 
{
    .size = 20,
    .radius = {128, 128, 128, 128},
    .inc    = {359,   1,  11,  20}
};

static void next_pixel_kreis (test_t*);
static void testbild (test_t*);
static void quadrate (test_t*);


static uint8_t 
onKey_test (test_t * t, uint8_t code)
{
    uint8_t id = t->id & 3;
    
    if (code == RC5_ROT)           id = 0;
    else if (code == RC5_GRUN)     id = 1;
    else if (code == RC5_GELB)     id = 2;
    else if (code == RC5_BLAU)     id = 3;
    
    uint8_t r = t->radius[id];
    uint8_t x = t->x0;
    uint8_t y = t->y0;
    
    if (code == RC5_LEFT)         r--,x-=10,t->size--;
    else if (code == RC5_RIGHT)   r++,x+=10,t->size++;
    else if (code == RC5_UP2)     r+=6,y+=10,t->a++;
    else if (code == RC5_DOWN2)   r-=6,y-=10,t->a--;
    
    t->x0 = x;
    t->y0 = y;
    
    if (r < 15) r = 15;
    if (r > 200) r = 200;

    t->radius[id] = r;

    int16_t inc = t->inc[id];    
    
    if (code == RC5_11)        inc--;
    else if (code == RC5_12)   inc++;
    else if (code == RC5_TV)   inc+=10;
    else if (code == RC5_PR)   inc-=6,t->mode^=1;
    
    if (inc <= 0)  inc = 1;
    if (inc >=720) inc = 719;

    t->inc[id] = inc;
    
    int16_t shift = t->shift[id];
    
    if (8 == code)  shift--;
    if (9 == code)  shift++;
    
    if (shift >= 720)   shift -= 720;
    if (shift <    0)   shift += 720;
    
    t->shift[id] = shift;
    
cli();    
    t->i = shift;
    if (id == 1)
        t->i = 0;
sei();
    t->id = id;
    
    return 1;
}

// Pixel pro Atom
#define PPA 12

static const int8_t atom_dxy[] PROGMEM = 
{
#if PPA == 12
    #define PPA_OFF 3
    //2, 2, 2, 1,  0, -1, -2, -2, -2, -1, 0, 1, 2, 2, 2
    1, 1, 1, 0, -1, -2, -2, -2, -2, -1, 0, 1, 1, 1, 1, 0
#elif PPA == 8
    #define PPA_OFF 2
    2, 2, 1, 0, -1, -1, 0, 1, 2, 2
#elif PPA == 4
    #define PPA_OFF 1
    1, 0, -1, 0, 1
#elif PPA == 1
    #define PPA_OFF 0
    0
#endif
};

//static pixel8_t atom_xy[120];

static uint8_t // NOINLINE
atom (test_t * t, pixel8_t a)
{
    uint8_t k = t->k;
    
    const int8_t * pdxy = & atom_dxy[k];
    uint8_t x = 0x80 + pgm_read_byte (pdxy);
    uint8_t y = 0x80 + pgm_read_byte (pdxy+PPA_OFF);
    
    beam_xy (a.x+x, a.y+y);
    
    if (!k)
        beam_skip();
    
    if (++k > PPA)
        k = 0;
        
    t->k = k;    
    
    return !k;
}

static const int8_t dxy[] PROGMEM =
{ 1, 1, 1, 0, -1, -1, -1, 0, 1, 1 };

#include "parith-15.h"

static pixel8_t NOINLINE
atom_move (pixel8_t p)
{        
    pixel8_t q;
    uint16_t d = p.x*p.x + p.y*p.y;

    uint8_t out = (d >= (uint16_t) 126*126);

    do
    {
        q = p;
        
        uint8_t rnd = 7 & prandom();
        const int8_t *pxy = & dxy[rnd];
    
        q.x += pgm_read_byte (pxy);
        q.y += pgm_read_byte (pxy+2);
        
        if (out)
            break;
            
        d = q.x*q.x + q.y*q.y;
    } while (d >= (uint16_t) 126*126);
    
    return q;
}

static inline uint16_t 
norm_720 (uint16_t un)
{
    int16_t n = un;
    
    do n -= 720;
    while (n >= 0);
    
    do n += 720;
    while (n < 0);
    
    return n;
}

static uint8_t NOINLINE
atome_kreis (test_t * t, uint16_t inc /* 720/num */, frac8_t r)
{
    RELOAD ("y", t);
    
    pixel8_t m;
    m.x = t->x0;
    m.y = t->y0;
    if (atom (t, m))
    {
        uint16_t i = inc + t->i2;
        
        if (i >= 720)
            i = 0;
            
        point8_t sc = _i_get_sincos8 (i);
        
        frotate8 (t->x0, t->y0, 0, r, sc);
            
        t->i2 = i;
        
        return !i;
    }
    
    return 0;
}

static void NOINLINE
atome_kreise (test_t * t)
{
    uint8_t c = t->cur;
    
    if (c == 0)
    {
        if (atome_kreis (t, 720/36, 124))            c = 1;
    }
    else if (c == 1)
    {
        if (atome_kreis (t, 720/24, 83))            c = 2;
    }
    else if (c == 2)
    {
        if (atome_kreis (t, 720/12, 42))            c = 3;
    }
    else
    {
        if (atome_kreis (t, 720, 0))            c = 0;
    }
    
    t->cur = c;
}

static void NOINLINE
atome (test_t * t)
{
    RELOAD ("y", t);
    
    if (atom (t, framebuf.test.atom_xy[t->j]))
    {
        uint8_t j = t->j;
        
        framebuf.test.atom_xy[j] = atom_move (framebuf.test.atom_xy[j]);
        
        if (++j >= NATOMS)
            j = 0;
            
        t->j = j;
    }
}

static uint8_t NOINLINE
quadrat (test_t * t)
{
    uint8_t j = t->j;
    
    uint8_t x, y;
    uint8_t size = t->size;
    uint8_t k = t->k;
    
    if (!j)
    {
        x = t->x0 - size;
        y = t->y0 + size;
        k = 0;
        
        beam_skip();
    }
    else
    {
        x = t->x;
        y = t->y;
        
        if (k == 0) x++;
        if (k == 1) y--;
        if (k == 2) x--;
        if (k == 3) y++;
    }
    
    beam_xy (x, y);
    t->x = x;
    t->y = y;
    
    if (++j > 2*size)
    {
        j = 1;
        
        if (++k >= 4)
            return 1;
    }
    
    t->k = k;
    t->j = j;
    
    return 0;
}

static void NOINLINE 
quadrate (test_t * t)
{
    if (quadrat (t))
    {
        static uint16_t sci;
        uint8_t q = 1 + t->q;
        int8_t x0=0, y0=0;
        int16_t sci2 = sci;
        
        if (q == 9)
        {
            if (++sci2 >= 720)
                sci2 = 0;
            q = 0;
            sci = sci2;
        }    
            
        int8_t pos = 105; // ~ sqrt{128-20} - 20
        if (q)
        {
            x0 = pos;
            sci2 += (uint16_t) 90*q;
            if (sci2 >= 720)
                sci2 -= 720;
            if (q & 1)
                sci2 = 720-sci2;
        }

        point8_t sc = _i_get_sincos8 (sci2);
        frotate8 (x0, y0, x0, y0, sc);
            
        t->q = q;
        t->j = 0;
        t->x0 = x0+0x80;
        t->y0 = y0+0x80;
    }

    return;
    //////////////////////////////
    uint16_t i = t->i;
    uint8_t k = t->k;
    int8_t x = i;
    int8_t y = 0;
    int8_t h;
    int8_t k3 = k & 3;
    
    x -= 50;
    y += 50;
    
    if (k3 > 1)
        x = -x, y = -y;
        
    if (k3 & 1)
        h = -x, x = y, y = h;
        
    beam_xy (x+t->x0, y+t->y0);
    
    if (++k == 5)
    {
        if (++i == 101)
        {
            i = 0;
            t-> x0 = 0x80;
            t-> y0 = 0x80;
        }
        t->i = i;    
        k = 0;
    }
    t->k = k;    
}

void 
testbild (test_t * t)
{
    uint16_t i = t->i;
    static uint8_t j;
    
    uint8_t x = i;
    uint8_t y = i >> 8;
    
    if (j >= 2)     y = ~y;
    if (j & 1)      y ^= x, x ^= y, y ^= x;
    
    beam_xy (x, y);
    
    if (!++i)
        j = (j+1) & 3;
        
    t->i = i;    
}

static void 
onPixel_test (test_t * t, void * tf UNUSED)
{
    FLAGS |= FRAME_LOCKED;    
    RELOAD ("y", t);
    if (t->id < 1)
        next_pixel_kreis (t);
    else if (t->id == 1)
        atome (t);
    else if (t->id == 2)
        quadrate (t);
    else
        //atome(t);//testbild (t);
        atome_kreise (t);
        
}

static void 
onEnter_test (test_t * t UNUSED)
{
    FLAGS |= FRAME_VALID;
}

void 
next_pixel_kreis (test_t * t)
{
    uint16_t i = t->i;
    uint8_t id = t->id;
    
    if (t != & test)
        return testbild(t);
    
    i += t->inc[id];
    if (i > 720)        i -= 720;
    if (i == 720)       i = 0;
    
    t->i = i;
    
    uint8_t x, y;
    
    point16_t sincos = _i_get_sincos (i);
    
    frac16_t radius = (uint16_t) t->radius[id] << 8;
    
    frac16_t px = _i_fmul16_16 (radius, sincos.x);
    frac16_t py = _i_fmul16_16 (radius, sincos.y);
    
    x = F16_TO_VAL (px);
    y = F16_TO_VAL (py);

    beam_xy (x, y);
}

const menuitem_t menuitem_testbild PROGMEM =
{
    .onPixel     = onPixel_test,
    .onEnter     = onEnter_test,
    .onKey       = onKey_test,
    .onRedraw    = (void*) 0,
    .redrawOn    = REDRAW_KEY,// | REDRAW_10_MILLI_SECONDS,
    .bufsize     = 0,
    .data        = & test
};


#if 0
void next_pixel_e(void)
{
    static uint16_t m = 3;
    const uint16_t M = 32767-2*kant;

    uint16_t a = m+m;
    if (a >= M) a -= M;

    a += m;
    if (a >= M) a -= M;
    m = a;
    
    uint8_t x = a & 127;
    uint8_t y = (a >> 7) & 127;
    push_xy (2*x, 2*y);
}

void next_pixel_b(void)
{
    static uint8_t x;
    static uint8_t y;
    const uint8_t A = kant;
    const uint8_t kant = 29;
    if (++x > kant)
    {
        x = 0;
        if (++y > kant)
            y = 0;
    }
    
    uint8_t ll = 0x80 - A*kant/2;
    push_xy (ll+A*x, ll+A*y);
}

uint8_t next_pixel_a (uint8_t strich)
{
    static uint8_t i;
    uint8_t ret = 0;
    
    if (kant < 30)  kant = 30;
    if (kant > 255-30) kant = 255-30;
    
    if (++i >= kant)
    {
        i = 0;
        if (++seit == 4)
        {
            seit = 0;
            ret = 1;
        }
    }

    if (strich)
    {
        if (((i+bord) & 15) == 0)    
        {
            SET (PORT_Z);
            i += 8;
            return 0;
        }
        else
            CLR (PORT_Z);
    }
    
    uint8_t k2 = kant/2;
    uint8_t k1 = 0x80+k2;
    uint8_t k0 = 0x80-k2;
    uint8_t ki0 = k0+i;
    uint8_t ki1 = k1-i;
    
    uint8_t x=0, y=0;
    
    if (0 == seit)
    {
        x = ki0;
        y = k0;
    }
    
    if (1 == seit)
    {
        x = k1;
        y = ki0;
    }
    
    if (2 == seit)
    {
        x = ki1;
        y = k1;
    }
    
    if (3 == seit)
    {
        x = k0;
        y = ki1;
    }
    
    push_xy (x, y);
    
    return ret;
}

void next_pixel_i(void)
{
    static uint8_t x;
    
    if (++x == 0)
    {
        push_xy (0,0);
    }
    
    if (porpo)
        push_xy (kant, x);
    else
        push_xy (x, kant);
}

void next_pixel_h(void)
{
    static uint8_t kreuz;
    static uint8_t index;
    
    uint8_t i = index;
    uint8_t k = kreuz;
    
    uint8_t len = kant;
    
    if (++i >= len)
    {
        i = 0;
        if (++k >= 2*9)
            k = 0;
    }
    
    kreuz = k;
    index = i;
    
    uint8_t x, y;
    uint8_t k2 = k/2;
    uint8_t l2 = len/2;
    
    if (k & 1)
    {
        y = i-l2;
        x = 0;
    }
    else
    {
        y = 0;
        x = i-l2;
    }
    
    if (k2==0)   x+=1+l2, y+=0x80;
    if (k2==1)   x+=0x80, y+=0xff-l2;
    if (k2==2)   x+=0xff-l2, y+=0x80;
    if (k2==3)   x+=0x80, y+=1+l2;
    if (k2==4)   x+=0x80, y+=0x80;
    
    if (k2 >= 5)
    {
        uint16_t l8_16 = 181*len; // 181 = 2^9 / sqrt(8)
        uint8_t l8 = l8_16 >> 9;   // len / sqrt(8)
        
        l8 = 90-l8; // 90 = 127 / sqrt(2)
        
        x += 0x80;
        y += 0x80;
        
        if (k2 == 5) x -= l8, y+= l8;
        if (k2 == 6) x += l8, y+= l8;
        if (k2 == 7) x -= l8, y-= l8;
        if (k2 == 8) x += l8, y-= l8;
    }
    
    
    push_xy (x, y);
}
#endif // 0
